/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl.arguments;

import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.AbstractLoader;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.Node;

/**
 * A delegating loader that replace argument references by their actual values.
 */
public class ArgumentLoader extends AbstractLoader {

  // --------------------------------------------------------------------------
  // Implementation of the Loader interface
  // --------------------------------------------------------------------------

  public Definition load(final String name, final Map<Object, Object> context)
      throws ADLException {
    final Definition d = clientLoader.load(name, context);
    final Map<String, Object> values = new HashMap<String, Object>();
    if (d instanceof ArgumentDefinition) {
      final String args = ((ArgumentDefinition) d).getArguments();
      if (args != null) {
        final StringTokenizer st = new StringTokenizer(args, ",");
        int i = 0;
        while (st.hasMoreTokens()) {
          final String arg = st.nextToken();
          String argname, argdef;
          final int j = arg.indexOf('=');
          if (j != -1) {
            argname = arg.substring(0, j);
            argdef = arg.substring(j + 1);
          } else {
            argname = arg;
            argdef = null;
          }
          Object argval = context.get(argname);
          if (argval == null) argval = argdef;
          if (argval == null) argval = context.get(" arg " + i);
          if (argval == null)
            throw new ADLException("No value defined for argument '" + argname
                + "'", (Node) d, null);
          values.put(argname, argval);
          ++i;
        }
        ((ArgumentDefinition) d).setArguments(null);
      }
    }
    evaluate((Node) d, values);
    return d;
  }

  // --------------------------------------------------------------------------
  // Argument evaluation methods
  // --------------------------------------------------------------------------

  void evaluate(final Node node, final Map<String, Object> context)
      throws ADLException {
    final Map<String, String> attrs = node.astGetAttributes();
    for (final Map.Entry<String, String> attrEntry : attrs.entrySet()) {
      final String attr = attrEntry.getKey();
      String value = attrEntry.getValue();
      if (value != null) {
        try {
          value = evaluate(value, context);
        } catch (final ADLException e) {
          throw new ADLException(e.getMessage(), node, null);
        }
        attrs.put(attr, value);
      }
    }
    node.astSetAttributes(attrs);

    for (final String nodeType : node.astGetNodeTypes()) {
      for (final Node n : node.astGetNodes(nodeType)) {
        if (n != null) {
          evaluate(n, context);
        }
      }
    }
  }

  // TODO created dedicated exception class.
  String evaluate(final String s, final Map<String, Object> context)
      throws ADLException {
    final int i = s.indexOf("${");
    if (i == -1) {
      return s;
    }
    final int j = s.indexOf('}', i);
    if (j == -1) {
      throw new ADLException("Malformed variable reference in '" + s + "'");
    }
    final String arg = s.substring(i + 2, j);
    final Object value = context.get(arg);
    if (value == null) {
      throw new ADLException("Undefined variable '" + arg + "'");
    }
    if (!(value instanceof String)) {
      throw new ADLException("Invalid variable value for argument '" + arg
          + "': not a String");
    }
    return evaluate(s.substring(0, i) + value + s.substring(j + 1), context);
  }
}
