/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2006 France Telecom R&D
 * Copyright (C) 2006-2007 INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Eric Bruneton
 *
 * Contributor: Philippe Merle, Matthieu Leclercq
 *
 * $Id: ComponentLoader.java 3624 2007-07-24 14:50:38Z apace $
 */

package org.objectweb.fractal.adl.components;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Type;
import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.AbstractLoader;
import org.objectweb.fractal.adl.AbstractNode;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.NodeClassLoader;
import org.objectweb.fractal.adl.util.ClassLoaderHelper;

/**
 * A {@link org.objectweb.fractal.adl.Loader} to check {@link Component} nodes
 * in definitions. This loader checks sub component names, loads the definitions
 * referenced by the component nodes (and the "extends" attribute), and merges
 * all these definitions into a single one (by following inheritance rules).
 */
public class ComponentLoader extends AbstractLoader
    implements
      ComponentLoaderAttributes {

  /**
   * The names of the "name" attribute for each AST node type. This map
   * associates the names of the "name" attribute, used to detect overridden
   * elements, to each AST node type that has such an attribute.
   */
  private final Map<String, String> nameAttributes;

  /**
   * The class loader used to merge AST node classes. This class merging is
   * necessary when merging several ASTs with different sources (eg different
   * DTDs).
   */
  private final MergeClassLoader    mergeClassLoader;

  // -------------------------------------------------------------------------
  // Constructor
  // -------------------------------------------------------------------------

  /**
   * Default constructor.
   */
  public ComponentLoader() {
    nameAttributes = new HashMap<String, String>();
    nameAttributes.put("component", "name");
    nameAttributes.put("interface", "name");
    nameAttributes.put("binding", "from");
    nameAttributes.put("attribute", "name");
    nameAttributes.put("coordinates", "name");
    // XXX old version commented, copied version from the trunk
    // mergeClassLoader = new
    // MergeClassLoader(ClassLoaderHelper.getClassLoader(this));
    mergeClassLoader = new MergeClassLoader(ClassLoaderHelper
        .getClassLoader(this));
  }

  // -------------------------------------------------------------------------
  // Implementation of the ComponentLoaderAttributes interface
  // -------------------------------------------------------------------------

  public String getNameAttributes() {
    final StringBuffer b = new StringBuffer();
    for (final Map.Entry<String, String> e : nameAttributes.entrySet()) {
      b.append(e.getKey());
      b.append(' ');
      b.append(e.getValue());
      b.append(' ');
    }
    return b.toString();
  }

  public void setNameAttributes(String nameAttributes) {
    this.nameAttributes.clear();
    String key = null;
    int p = nameAttributes.indexOf(' ');
    while (p != -1) {
      final String s = nameAttributes.substring(0, p);
      if (key == null) {
        key = s;
      } else {
        this.nameAttributes.put(key, s);
        key = null;
      }
      nameAttributes = nameAttributes.substring(p + 1);
      p = nameAttributes.indexOf(' ');
    }
    if (key != null) {
      this.nameAttributes.put(key, nameAttributes);
    }
  }

  // -------------------------------------------------------------------------
  // Implementation of the Loader interface
  // -------------------------------------------------------------------------

  public Definition load(final String name, final Map<Object, Object> context)
      throws ADLException {
    return load(new ArrayList<String>(), name, context);
  }

  protected Definition load(final List<String> loaded, final String name,
      final Map<Object, Object> context) throws ADLException {
    if (loaded.contains(name)) {
      throw new ADLException("Cycle in definition references: " + loaded);
    }
    final List<String> l = new ArrayList<String>(loaded);
    l.add(name);
    Definition d = clientLoader.load(name, context);
    if (d instanceof ComponentDefinition) {
      ComponentDefinition container = (ComponentDefinition) d;
      normalizeComponentContainer(container);
      resolveComponentContainer(l, container, container, context);
      if (container.getExtends() != null) {
        final List<String> defs = parseDefinitions(container.getExtends());
        try {
          d = (Definition) merge((Node) d, (Node) resolveDefinitions(l, defs,
              context));
        } catch (final ADLException e) {
          throw new ADLException("Cannot load super definition(s)",
              (Node) container, e);
        }
        container = (ComponentDefinition) d;
        container.setExtends(null);
      }
      resolveSharedComponentContainer(container, container, context);
    }
    return d;
  }

  // -------------------------------------------------------------------------
  // First pass: normalization
  // -------------------------------------------------------------------------

  protected void normalizeComponentContainer(final ComponentContainer container)
      throws ADLException {
    final Set<String> names = new HashSet<String>();
    for (final Component comp : container.getComponents()) {
      final String name = comp.getName();
      if (name == null) {
        throw new ADLException("Component name missing", (Node) comp);
      }
      if (!names.add(name)) {
        throw new ADLException("Duplicated component name '" + name + "'",
            (Node) comp);
      }
      normalizeComponentContainer(comp);
      /*
       * if (name.indexOf('/') != -1) { container.removeComponent(comp);
       * addComponent(container, comp); }
       */
    }
  }

  // -------------------------------------------------------------------------
  // Second pass: definition references resolution
  // -------------------------------------------------------------------------

  protected void resolveComponentContainer(final List<String> loaded,
      final ComponentContainer topLevelDefinition,
      final ComponentContainer container, final Map<Object, Object> context)
      throws ADLException {
    for (final Component comp : container.getComponents()) {
      resolveComponentContainer(loaded, topLevelDefinition, comp, context);
      final String definition = comp.getDefinition();
      if (definition != null) {
        final List<String> defs = parseDefinitions(definition);
        // shared components will be resolved by the
        // resolveSharedComponentContainer method.
        if (defs.size() != 1 || !isShared(defs.get(0))) {
          comp.setDefinition(null);
          ((Node) comp).astSetDecoration("definition", definition);
          Definition d;
          try {
            d = resolveDefinitions(loaded, defs, context);
          } catch (final ADLException e) {
            throw new ADLException("Cannot load referenced definition(s)",
                (Node) comp, e);
          }
          final Map<Node, Node> replacements = new HashMap<Node, Node>();
          merge((Node) comp, (Node) d, replacements);
          replaceComponents(topLevelDefinition, replacements);
        }
      }
    }
  }

  protected void resolveSharedComponentContainer(
      final ComponentContainer topLevelDefinition,
      final ComponentContainer container, final Map<Object, Object> context)
      throws ADLException {
    final Component[] comps = container.getComponents();
    for (int i = 0; i < comps.length; i++) {
      final Component comp = comps[i];
      resolveSharedComponentContainer(topLevelDefinition, comp, context);
      String definition = comp.getDefinition();
      if (definition != null) {
        final List<String> defs = parseDefinitions(definition);
        if (defs.size() == 1 && isShared(defs.get(0))) {
          // shared component
          comp.setDefinition(null);
          ((Node) comp).astSetDecoration("definition", definition);
          if (definition.startsWith("./")) {
            definition = definition.substring(2);
          }
          final Component c = getPathComponent(topLevelDefinition, definition);
          if (c == null) {
            throw new ADLException("No such component", (Node) comp);
          }
          if (!c.getName().equals(comps[i].getName())) {
            throw new ADLException(
                "Shared components with distinct names not yet supported",
                (Node) comp);
          }
          final Map<Node, Node> replacements = new HashMap<Node, Node>();
          replacements.put((Node) comp, (Node) c);
          replaceComponents(topLevelDefinition, replacements);
        }
      }
    }
  }

  protected List<String> parseDefinitions(String nameList) {
    final List<String> l = new ArrayList<String>();
    int p = nameList.indexOf(',');
    while (p != -1) {
      l.add(nameList.substring(0, p));
      nameList = nameList.substring(p + 1);
      p = nameList.indexOf(',');
    }
    l.add(nameList);
    return l;
  }

  protected boolean isShared(final String definition) {
    return definition.indexOf('/') != -1;
  }

  protected Definition resolveDefinitions(final List<String> loaded,
      final List<String> nameList, final Map<Object, Object> context)
      throws ADLException {
    Definition d = load(loaded, nameList.get(0), context);
    for (int i = 1; i < nameList.size(); ++i) {
      final Definition e = load(loaded, nameList.get(i), context);
      d = (Definition) merge((Node) e, (Node) d);
    }
    return d;
  }

  protected Component getComponent(final ComponentContainer container,
      final String name) {
    for (final Component comp : container.getComponents()) {
      if (comp.getName().equals(name)) {
        return comp;
      }
    }
    return null;
  }

  protected Component getPathComponent(final ComponentContainer container,
      final String name) {
    final int p = name.indexOf('/');
    if (p == -1) {
      return getComponent(container, name);
    }
    final Component parent = getComponent(container, name.substring(0, p));
    if (parent == null) return null;
    return getPathComponent(parent, name.substring(p + 1));
  }

  protected ComponentContainer replaceComponents(
      final ComponentContainer container, final Map<Node, Node> replacements) {
    final Node node = replacements.get(container);
    if (node != null) {
      return (ComponentContainer) node;
    }
    final Component[] comps = container.getComponents();
    for (final Component element : comps) {
      container.removeComponent(element);
    }
    for (final Component element : comps) {
      container.addComponent((Component) replaceComponents(element,
          replacements));
    }
    return container;
  }

  // -------------------------------------------------------------------------
  // Utility methods: inheritance resolution
  // -------------------------------------------------------------------------

  protected Node merge(final Node elem, final Node superElem)
      throws ADLException {
    return merge(elem, superElem, null);
  }

  protected Node merge(final Node elem, final Node superElem,
      final Map<Node, Node> replacements) throws ADLException {
    final Map<Node, MergeInfo> infos = new HashMap<Node, MergeInfo>();

    // compute merge info
    final MergeInfo elemInfo = computeMergeInfos(elem, superElem, infos);

    // create the merged nodes
    for (final MergeInfo info : infos.values()) {
      info.createResultNode(mergeClassLoader);
    }

    // initialize the merged nodes
    final Node n = initMergedNodes(elemInfo, infos);

    if (replacements != null) {
      for (final Map.Entry<Node, MergeInfo> e : infos.entrySet()) {
        replacements.put(e.getKey(), e.getValue().result);
      }
    }
    return n;
  }

  private MergeInfo computeMergeInfos(final Node elem, final Node superElem,
      final Map<Node, MergeInfo> infos) throws ADLException {
    MergeInfo info = infos.get(elem);
    if (info == null) {
      info = infos.get(superElem);
      if (info != null) {
        // 'elem' is not known but 'superElem' is (i.e. 'superElem' is shared)
        // this means that 'elem' is in fact shared with 'info.elem'
        info.resetElem(elem);
      } else { // info == null
        info = new MergeInfo(elem);
        info.addSuperNode(superElem);
        infos.put(superElem, info);
      }
      infos.put(elem, info);
    } else if (info.addSuperNode(superElem)) {
      // superNode added
      infos.put(superElem, info);
    } else {
      // superNode already in the set of the superNodes; it has already been
      // merged, simply return
      return info;
    }

    // build the set of the type of the sub nodes.
    final Set<String> nodeTypes = new HashSet<String>();
    nodeTypes.addAll(Arrays.asList(elem.astGetNodeTypes()));
    nodeTypes.addAll(Arrays.asList(superElem.astGetNodeTypes()));

    // for each type of sub nodes
    for (final String nodeType : nodeTypes) {
      // get the sub nodes of 'superNode' of type 'nodeType'
      Node[] superNodes = superElem.astGetNodes(nodeType);
      // get the sub nodes of 'elem' of type 'nodeType'
      Node[] nodes = elem.astGetNodes(nodeType);

      // if there is no node to merge, continue with the next 'nodeType'.
      if ((superNodes == null || superNodes.length == 0)
          && (nodes == null || nodes.length == 0)) continue;

      if (superNodes == null) superNodes = new Node[0];
      if (nodes == null) nodes = new Node[0];

      // check and get the arity of 'nodeType'
      final SubNodeArity arity = checkArityCompatibility(elem, superElem,
          nodeType);

      if (arity == SubNodeArity.ONE) {
        // arity of sub nodes is 'ONE'. check that there is not more that one
        // node in 'superNodes' and in 'Nodes'
        assert superNodes.length <= 1;
        assert nodes.length <= 1;

        if (superNodes.length > 0 && superNodes[0] != null) {
          // there is a node in 'superNodes'
          if (nodes.length == 0 || nodes[0] == null) {
            // there is no node in 'nodes'
            info.addSubNodeInfo(computeMergeInfos(superNodes[0], infos));
          } else {
            // there is a node both in 'nodes' and 'superNodes'. merge them
            info.addSubNodeInfo(computeMergeInfos(nodes[0], superNodes[0],
                infos));
          }
        } else if (nodes.length > 0 && nodes[0] != null) {
          // there is a node only in 'nodes'
          info.addSubNodeInfo(computeMergeInfos(nodes[0], infos));
        }
      } else { // arity == SubNodeArity.MANY
        // get the name of the attribute that identify the one sub node of type
        // 'nodeType' among others.
        final String nameAttr = nameAttributes.get(nodeType);
        if (nameAttr == null) {
          // there is no attribute that identify sub nodes, do not merge nodes
          for (final Node n : superNodes)
            info.addSubNodeInfo(computeMergeInfos(n, infos));
          for (final Node n : nodes)
            info.addSubNodeInfo(computeMergeInfos(n, infos));

        } else { // nameAttr != null
          // there is an attribute that identify sub nodes, find a node of
          // 'nodes' and a node of 'superNodes' that have the same value for
          // this attribute.
          superNodeLoop : for (final Node superNode : superNodes) {
            // get the value of attribute 'nameAttr' of node 'superNode'
            final String superName = superNode.astGetAttributes().get(nameAttr);
            // Check that the value is not null
            if (superName == null)
              throw new ADLException("The attribute '" + nameAttr
                  + "' cannot be null", superNode);

            // for each node in 'Nodes'
            for (final Node node : nodes) {
              // get the value of attribute 'nameAttr' of node 'node'
              final String name = node.astGetAttributes().get(nameAttr);
              // Check that the value is not null
              if (name == null)
                throw new ADLException("The attribute '" + nameAttr
                    + "' cannot be null", node);

              if (name.equals(superName)) {
                // 'name' and 'superName' are equals, merge 'node' and
                // 'superNode'
                info.addSubNodeInfo(computeMergeInfos(node, superNode, infos));
                // then continue with the next 'superNode'
                continue superNodeLoop;
              }
            }
            // no node in 'nodes' has the same name as 'superName'
            // superNode is not merged
            info.addSubNodeInfo(computeMergeInfos(superNode, infos));
          }

          // find node in 'nodes' that are not merged with a node in
          // 'superNodes'
          nodeLoop : for (final Node node : nodes) {
            // get the value of attribute 'nameAttr' of node 'node'
            final String name = node.astGetAttributes().get(nameAttr);
            for (final Node superNode : superNodes) {
              // get the value of attribute 'nameAttr' of node 'superNode'
              final String superName = superNode.astGetAttributes().get(
                  nameAttr);
              if (name.equals(superName)) {
                // 'node' and 'superNode' has the same name, they are merged
                // skip and continue with the next 'node'
                continue nodeLoop;
              }
            }

            // no node in 'superNodes' has the same name as 'name'
            // node is not merged.
            info.addSubNodeInfo(computeMergeInfos(node, infos));
          }
        }
      }
    }
    return info;
  }

  private MergeInfo computeMergeInfos(final Node node,
      final Map<Node, MergeInfo> infos) throws ADLException {
    MergeInfo info = infos.get(node);
    if (info == null) {
      info = new MergeInfo(node);
      infos.put(node, info);
      for (final String type : node.astGetNodeTypes()) {
        for (final Node subNode : node.astGetNodes(type)) {
          if (subNode != null) {
            info.addSubNodeInfo(computeMergeInfos(subNode, infos));
          }
        }
      }
    }
    return info;
  }

  private Node initMergedNodes(final MergeInfo info,
      final Map<Node, MergeInfo> infos) {
    if (info.done) {
      return info.result;
    }
    info.done = true;

    final Node elem = info.elem;
    final Node result = info.result;

    // merges source
    result.astSetSource(elem.astGetSource());

    // merges attributes
    final Map<String, String> resultAttrs = elem.astGetAttributes();

    // for each node that is overridden by 'elem'
    for (final Node superNode : info.getSuperNodes()) {
      // for each attribute of 'superNode'
      for (final Map.Entry<String, String> entry : superNode.astGetAttributes()
          .entrySet()) {
        final String name = entry.getKey();
        final String superValue = entry.getValue();
        final String value = resultAttrs.get(name);
        if (superValue != null && value == null) {
          // the attribute is defined for 'superNode' and not for 'elem'
          // put the value from 'superNode' as value of the attribute.
          resultAttrs.put(name, superValue);
        }
      }
    }
    result.astSetAttributes(resultAttrs);

    // merges decorations
    // decorations are merged from the "less significant" node to the most one.
    final Map<String, Object> resultDecors = elem.astGetDecorations();
    for (final Node superNode : info.getSuperNodes()) {
      // for each decoration of 'superNode'
      for (final Map.Entry<String, Object> entry : superNode
          .astGetDecorations().entrySet()) {
        final String name = entry.getKey();
        final Object superValue = entry.getValue();
        final Object value = resultDecors.get(name);
        if (superValue != null && value == null) {
          // the decoration is defined for 'superNode' and not for 'elem'
          // put the value from 'superNode' as value of the decoration.
          resultDecors.put(name, superValue);
        }
      }
    }
    result.astSetDecorations(resultDecors);

    // merges sub nodes

    for (final MergeInfo subNodeInfo : info.getSubNodeInfos()) {
      result.astAddNode(initMergedNodes(subNodeInfo, infos));
    }
    return info.result;
  }

  /** The Arity of a sub node. */
  private enum SubNodeArity {
    /** one sub node. */
    ONE,
    /** many sub nodes. */
    MANY,
    /** unknown. */
    NONE
  }

  private final Map<Class<? extends Node>, Map<String, SubNodeArity>> subNodeArityCache = new IdentityHashMap<Class<? extends Node>, Map<String, SubNodeArity>>();

  private SubNodeArity getSubNodeArity(final Node node, final String subNodeType) {
    final Class<? extends Node> nodeClass = node.getClass();
    // try to get arities for the 'nodeClass' in the cache
    Map<String, SubNodeArity> arities = subNodeArityCache.get(nodeClass);
    if (arities == null) {
      // arities not found in cache for 'nodeClass'.
      // create a new Map and add it in cache.
      arities = new HashMap<String, SubNodeArity>();
      subNodeArityCache.put(nodeClass, arities);
    }

    // try to get arity if 'subNodeType' in map of arities of 'nodeClass'
    SubNodeArity arity = arities.get(subNodeType);
    if (arity == null) {
      // not found in cache.
      // try to find the getter method of 'subNodeType' as if the arity is 'ONE'
      // if the method is found, the arity is 'ONE'.
      try {
        nodeClass.getMethod("get"
            + Character.toUpperCase(subNodeType.charAt(0))
            + subNodeType.substring(1), new Class[0]);
        arity = SubNodeArity.ONE;
      } catch (final Exception e) {
        // method for arity ONE not found try for arity 'MANY,
        try {
          nodeClass.getMethod("get"
              + Character.toUpperCase(subNodeType.charAt(0))
              + subNodeType.substring(1) + 's', new Class[0]);
          arity = SubNodeArity.MANY;
        } catch (final Exception e1) {
          // not found, this node class does not provides method for sub node of
          // type 'subNodeType'
          arity = SubNodeArity.NONE;
        }
      }
      // put the arity in cache.
      arities.put(subNodeType, arity);
    }

    return arity;
  }

  private SubNodeArity checkArityCompatibility(final Node node,
      final Node superNode, final String subNodeType) throws ADLException {
    final SubNodeArity nodeArity = getSubNodeArity(node, subNodeType);

    // if node and superNode are instances of the same class, they have the same
    // arity for 'subNodeType'
    if (node.getClass() == superNode.getClass()) {
      assert nodeArity != SubNodeArity.NONE;
      return nodeArity;
    }

    final SubNodeArity superNodeArity = getSubNodeArity(superNode, subNodeType);

    if (nodeArity == SubNodeArity.NONE) {
      assert superNodeArity != SubNodeArity.NONE;
      return superNodeArity;
    }

    if (superNodeArity == SubNodeArity.NONE) {
      return nodeArity;
    }

    if (nodeArity != superNodeArity) {
      throw new ADLException(
          "Cannot merge AST, the arity of sub nodes of type '" + subNodeType
              + "' is not the same for node '" + node.astGetSource()
              + "' and node '" + superNode.astGetSource() + "'. Check DTDs");
    }
    return nodeArity;
  }

  private static class MergeInfo {
    // the node that must be merged with the 'superNodes'.
    Node            elem;

    // the set of the nodes that must be merged with 'elem' to obtain the
    // 'result' node.
    Set<Node>       superNodes;

    // the Node that is the result of the merge of the 'nodes'.
    Node            result;

    // the list of merge info of the sub nodes
    List<MergeInfo> subNodeInfos    = new ArrayList<MergeInfo>();

    // true if the merge has been done
    boolean         done;

    // true if every 'superNodes' are instances of the same class as 'elem'
    boolean         uniqueNodeClass = true;

    MergeInfo(final Node elem) {
      assert elem != null;
      this.elem = elem;
    }

    // This method is used if 2 elements override the same superElement (i.e. in
    // case of shared component).
    void resetElem(final Node newElem) {
      assert newElem != null;
      assert elem != null;
      assert superNodes != null;
      final Node oldElem = elem;
      elem = newElem;
      final Set<Node> oldSuperNodes = superNodes;
      superNodes = new LinkedHashSet<Node>();
      superNodes.add(oldElem);
      superNodes.addAll(oldSuperNodes);
    }

    boolean addSuperNode(final Node superNode) {
      assert superNode != null;
      if (superNodes == null) {
        superNodes = new LinkedHashSet<Node>();
      }
      // check 'superNode' class and update 'uniqueNodeClass'
      if (superNode.getClass() != elem.getClass()) uniqueNodeClass = false;
      return superNodes.add(superNode);
    }

    @SuppressWarnings("unchecked")
    Set<Node> getSuperNodes() {
      if (superNodes == null) return Collections.EMPTY_SET;
      return superNodes;
    }

    void addSubNodeInfo(final MergeInfo subNodeInfo) {
      assert subNodeInfo != null;
      if (subNodeInfos == null) subNodeInfos = new ArrayList<MergeInfo>();
      subNodeInfos.add(subNodeInfo);
    }

    @SuppressWarnings("unchecked")
    List<MergeInfo> getSubNodeInfos() {
      if (subNodeInfos == null) return Collections.EMPTY_LIST;
      return subNodeInfos;
    }

    void createResultNode(final MergeClassLoader mergeClassLoader)
        throws ADLException {
      if (uniqueNodeClass) {
        try {
          result = elem.getClass().newInstance();
        } catch (final Exception e) {
          throw new ADLException("Cannot merge ASTs", elem, e);
        }
      } else {
        final Class<? extends Node> elemClass = elem.getClass();
        int hash = elemClass.hashCode();

        final List<Class<?>> nodeClasses = new ArrayList<Class<?>>(superNodes
            .size() + 1);
        nodeClasses.add(elemClass);
        for (final Node superNode : superNodes) {
          final Class<? extends Node> superNodeClass = superNode.getClass();
          hash = 17 * (hash + superNodeClass.hashCode());
          nodeClasses.add(superNodeClass);
        }
        final String code = Integer.toHexString(hash);
        final String name = "org.objectweb.fractal.adl.merged.Merged" + code;

        try {
          final Class<?> merged = mergeClassLoader.mergeClass(name, elem
              .astGetType(), nodeClasses);
          result = (Node) merged.newInstance();
        } catch (final Exception e) {
          throw new ADLException("Cannot merge AST classes", elem, e);
        }
      }
    }
  }

  private static class MergeClassLoader extends NodeClassLoader {

    MergeClassLoader(final ClassLoader parent) {
      super(parent);
    }

    Class<?> mergeClass(final String name, final String astNodeName,
        final List<Class<?>> classes) throws ClassNotFoundException {
      try {
        return loadClass(name);
      } catch (final ClassNotFoundException e) {
      }

      final Set<String> itfs = new HashSet<String>();
      for (final Class<?> c : classes) {
        for (final Class<?> itf : c.getInterfaces()) {
          itfs.add(Type.getInternalName(itf));
        }
      }

      final ClassWriter cw = generateClass(name, astNodeName, Type
          .getInternalName(AbstractNode.class), itfs.toArray(new String[itfs
          .size()]));

      return defineClass(name, cw.toByteArray());
    }
  }
}
