/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 *
 * Contributor(s): Matthieu Leclercq (externalize AST node factory)

 */

package org.objectweb.fractal.adl.xml;

import java.io.InputStream;
import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.ADLNotFoundException;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.Loader;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.Parser;
import org.objectweb.fractal.adl.ParserException;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

/**
 * A {@link Loader} that loads definitions from XML files.
 */

public class XMLLoader implements Loader, BindingController {

	private boolean validate = true;

	private Parser parser;

	public final static String NODE_FACTORY_BINDING = "node-factory";

	// --------------------------------------------------------------------------
	// Client interfaces
	// --------------------------------------------------------------------------

	/** The {@link XMLNodeFactory} client interface of this component. */

	public XMLNodeFactory nodeFactoryItf;

	/**
	 * Creates a new {@link XMLLoader}.
	 */
	public XMLLoader() {
	}

	/**
	 * Creates a new {@link XMLLoader}.
	 * 
	 * @param validate
	 *            if <code>true</code>, use a validating XML parser.
	 */
	public XMLLoader(final boolean validate) {
		this.validate = validate;
	}

	// --------------------------------------------------------------------------
	// Implementation of the Loader interface
	// --------------------------------------------------------------------------

	public Definition load(final String name, final Map context)
			throws ADLException {
		synchronized (this) {
			if (parser == null) {
				parser = new XMLParser(validate, nodeFactoryItf);
			}
		}

		try {
			final String file = name.replace('.', '/') + ".fractal";
			ClassLoader cl = null;
			if (context != null) {
				cl = (ClassLoader) context.get("classloader");
			}
			if (cl == null) {
				cl = getClass().getClassLoader();
			}
			final InputStream is = cl.getResourceAsStream(file);
			if (is == null) {
				throw new ADLNotFoundException("Cannot find '" + file
						+ "' in the classpath");
			}
			final Definition d = (Definition) parser.parse(is, file);
			if (d.getName() == null) {
				throw new ADLException("Definition name missing", (Node) d);
			}
			if (!d.getName().equals(name)) {
				throw new ADLException("Wrong definition name ('" + name
						+ "' expected, instead of '" + d.getName() + "')",
						(Node) d);
			}
			return d;
		} catch (final ParserException e) {
			throw new ADLException("Cannot load '" + name + "'", null, e);
		}
	}

	public void bindFc(String s, Object o) throws NoSuchInterfaceException,
			IllegalBindingException, IllegalLifeCycleException {
		if (NODE_FACTORY_BINDING.equals(s)) {
			this.nodeFactoryItf = (XMLNodeFactory) o;
		}

	}

	public String[] listFc() {
		return new String[] { NODE_FACTORY_BINDING };
	}

	public Object lookupFc(String s) throws NoSuchInterfaceException {
		if (NODE_FACTORY_BINDING.equals(s)) {
			return this.nodeFactoryItf;
		}
		return null;
	}

	public void unbindFc(String s) throws NoSuchInterfaceException,
			IllegalBindingException, IllegalLifeCycleException {
		if (NODE_FACTORY_BINDING.equals(s)) {
			this.nodeFactoryItf = null;
		}

	}
}
